%###############   MJC LAUNCHER  #######################
%
%
% ****   MAIN MODULE   ****
%


:- module(mjc, [
	      % --- prove and produce a latex file
	      prove/1,  %  prove(+FormLabel)
	      prove/2, %  prove(+FormLabel,+TestFile)
	     % --- prove and produce  latex and pdf  files
	      provepdf/1, % provepdf(+FormLabel)
	      provepdf/2, %  provepdf(+FormLabel,+TestFile)
	     % --- run the above predicates for all the defined formulas
	      prove_all/0, %
	      prove_all/1, %   prove_all(+TestFile)
	      provepdf_all/0, %
	      provepdf_all/1 % provepdf_all(+TestFile)
	  ]).

:- reexport(syntax).


:- set_prolog_flag(toplevel_print_options, [max_depth(200)]).

:- discontiguous test/2.  % suppress warning messagges
:- discontiguous test/3.

:- use_module(syntax).

:-use_module(proverMJCR,[
		 buildProof/3 % buildProof(+Seq,-Result,-ProofType)
	     ]
	    ).

:- use_module(printUtil,[
		 proof_to_latex/5,
		 generate_pdf/2,
		 print_error/1,
		print_error/2
	     ]).

:- multifile(test_dir/1).
:- multifile(test/3).
:- multifile(test/2).


default_test_file('test.pl').   %% default test file
default_test_dir('examples').   %% default test dir

%#############    MAIN PREDICATES  #######################

/*

In the following predicates:

  -  FormLabel:  the label of the  formula to be proved
   - TestFile:   the file containing the definition of the formula
		  If not specified, the default test file is used
                 (see  predicate default_test_file/1).

Both parameters must be Prolog atoms.

*  Prove and generate LaTeX file:

     prove(+FormLabel)
     prove(+FormLabel,+TestFile)

*  Prove and generate LaTeX and pdf files:

     provepdf(+FormLabel)
     provepdf(+FormLabel,+TestFile)

* Run the previous predicates for all the defined formulas:

    prove_all   // run prove(FormLabel)
                // for every FormLabel in the default test file
    prove_all(+TestFile)    // run prove(FormLabel,TestFile)
			    // for every FormLabel in TestFile
    provepdf_all
    provepdf_all(+TestFile)


*/


prove(FormLabel) :-
         default_test_file(TestFile),
	 prove(FormLabel,TestFile).


prove(FormLabel,TestFile) :-
	load_test_file(TestFile,TestFilePath),
	prove1(verbose,FormLabel,TestFilePath,ResultType,LatexFilePath),
	(
	  ResultType = proof  ->
	  writeln('** To compile:'),
	 ansi_format([bold,fg(green)], 'pdflatex  ~s\n', [LatexFilePath])
	    ;
	    true
	).

provepdf(FormLabel) :-
         default_test_file(TestFile),
	 provepdf(FormLabel,TestFile).

provepdf(FormLabel,TestFile)  :-
	load_test_file(TestFile,TestFilePath),
	provepdf1(verbose,TestFilePath,FormLabel).


prove_all :-
	 default_test_file(TestFile),
	 prove_all(TestFile).

prove_all(TestFile) :-
	load_test_file(TestFile,TestFilePath),
	findall(FormLabel, (test(FormLabel,_) ; test(FormLabel,_,_)), FormLabels),
	forall(member(FormLabel, FormLabels),	prove1(nonVerbose,FormLabel,TestFilePath,_ResultType,_LatexFilePath)),
	nl,writeln('**  See the LaTeX files (to be compiled with pdflatex) in:'),
	get_test_dir(Test_dir),
	ansi_format([bold,fg(green)], '~s\n', [Test_dir]).


provepdf_all :-
	 default_test_file(TestFile),
	 provepdf_all(TestFile).


provepdf_all(TestFile) :-
	load_test_file(TestFile,TestFilePath),
	findall(FormLabel, (test(FormLabel,_) ; test(FormLabel,_,_)), FormLabels),
	forall(member(FormLabel, FormLabels),	provepdf1(nonVerbose,TestFilePath,FormLabel)),
	nl,writeln('** See the LaTeX and pdf files in:'),
	get_test_dir(Test_dir),
        ansi_format([bold,fg(green)], '~s\n', [Test_dir]).


%% ######  AUXILIARY PREDICATES ###


% load_test_file(+TestFile,-TestFilePath)
% TestFile: the name of the  test file (relative or absolute path)
% TestPath: absolute name (path) of TestFile


load_test_file(TestFile,TestFilePath) :-
	absolute_file_name(TestFile,TestFilePath),
	(
	    exists_file(TestFilePath) ->
                   consult(TestFilePath)
              ;
             print_error('Test file not found:\n'),
	     print_error(' ~w\n', [TestFilePath]),
             abort
	).





% prove1(+Verbose,+FormLabel,+TestFile,-ResultType,-LatexFilePath)
%  proofs are written  in a latex file
%  Verbose: if verbose, prints more messages
%  FormLabel: the label of the  formula to be proved
%  TestFile: the current test file
%  LatexFilePath: the absolute path of the latex file
%  ResultType: proof if the formula is provable, ref otherwise
%  containing the derivations
%


prove1(Verbose,FormLabel,TestFile,ResultType,LatexFilePath) :-
	 (
	  get_formula(FormLabel,TestFile,Form ,DefsNF) ->
                 true
           ;
             % SYNTAX ERRORS
            print_error('Check the syntax of formulas (e.g., do not use \'->\' for implication, but \'-->\')'),
            abort
         ),
	  Seq = [] ==> Form @ [],  %% the MJCR sequent to be proved
	  ansi_format([bold,fg(cyan)], '**** FORMULA  \'~s\' ****\n', [FormLabel]),
	  writeln(Form),
	  buildProof(Seq,Result,ResultType),
	%  nl,writeln(Result),nl,
	  ( ResultType = proof ->
	       ansi_format([bold,fg(cyan)], '-- PROVABLE --\n', []),
	       get_test_dir(Test_dir),
	       atomic_list_concat(['proof_',FormLabel],BaseName),
	       file_name_extension(BaseName, tex, LatexFileName),
	       directory_file_path(Test_dir,LatexFileName,LatexFilePath1),
	       absolute_file_name( LatexFilePath1, LatexFilePath),
	       proof_to_latex(FormLabel,Form,Result,DefsNF,LatexFilePath), !,
	      (	   Verbose = verbose ->
	         writeln('** A derivation has been written in the LaTeX file:'),
	         ansi_format([bold,fg(green)], '~s\n', [LatexFilePath])
	      ;
	         % not verbose
                  true
	      )
	  ;
	  % FormLabel is not provable
	    ansi_format([bold,fg(cyan)], '-- NOT PROVABLE --\n', []),
	    format('** Interpretation I falsifying \'~s\' (list of true prop. variables):\n',[FormLabel]),
	    Result = ref(PropList),
	    format('   ~w\n', [PropList])
	  ).

provepdf1(Verbose,TestFile,FormLabel) :-
	prove1(Verbose,FormLabel,TestFile,ResultType,LatexFilePath),
	 ( ResultType = proof ->
	     generate_pdf(LatexFilePath,PdfFilePath),
	     (	  Verbose = verbose ->
	           writeln('** To display it, open the pdf file:'),
		  ansi_format([bold,fg(green)], '~s\n', [PdfFilePath])
	     ;
	         % not verbose
	         true
	     )
	% FormLabel is not provable
	 ;
	 true).






%% get_formula(+FormLabel, +TestFile,-Form,-DefsNF)

get_formula(FormLabel,TestFile,Form,DefsNF) :-
	test(FormLabel, F)  ->
	%% formula without definitions
	      Form = F,
	      replace_defs(F,[],Form,DefsNF) %check
	;
	test(FormLabel, F , Defs) ->
	%% formula with definitions
	    replace_defs(F,Defs,Form,DefsNF)
	;
	  %% SeqLabel is not a valid label
         print_error('No formula with label\n'),
	  print_error('  ~w\n', [FormLabel]),
	  print_error('is defined in file\n'),
	  print_error('  ~w\n', [TestFile]),
	  abort.



%   get_test_dir(-Test_dirPath)
%   Get the path of dir. containng  tex/pdf files

get_test_dir(Test_dirPath) :-
	(
              test_dir(Test_dir) ->
                   true
                   ;
               default_test_dir(Test_dir)
        ),
         (
	exists_directory(Test_dir) ->  % create test_dir if it does not exist
	true
	;
	make_directory(Test_dir)
	),
       absolute_file_name(Test_dir,Test_dirPath).



%###############   REPLACEMENTS	 ###############################


% replace_defs(+F,+Defs,-Form,-DefsNF)
%  F: a formula
%  Defs:  a list of definitions of the kind
%
%   a_1 = form_1 ,   .....   ,  a_n =  form_n
%
% where a_k's are Prolog atoms an form_k's are formulas
%
% Form is obtained by  performing the replacements in Defs
%
% We assume that definitions are not circular.
% For instance, definition of the form
%
%   a =  p & a
%
% are not allowed
%
%
%  A definition
%
%    a = form
%
% in in normal form iff form does not contain definitions.
%
% DefsNF contains the definitions of Defs in normal form


replace_defs(F,Defs,Form,DefsNF) :-
	set_defs_in_mormal_form(Defs,[], DefsNF),
	replace_defs_form(F,DefsNF,Form).

% set_defs_in_mormal_form(+Defs_todo, +DefsNF_done, -DesfNF)

set_defs_in_mormal_form([F = Df | Defs_todo], DefsNF_done,  DefsNF) :-
	union(DefsNF_done,Defs_todo, AllDefs),
	replace_defs_form(Df, AllDefs, DfNF),
	set_defs_in_mormal_form( Defs_todo, [ F = DfNF |  DefsNF_done]  ,   DefsNF).

set_defs_in_mormal_form([], DefsNF, DefsNF_rev) :-
	reverse(DefsNF,DefsNF_rev).



%%%   ERROR
replace_defs_form(F,_Defs,_Form) :-
	\+ ground(F), !,
	print_error('Not a valid formula (check the syntax):\n'),
	print_error(' ~w\n',[F]),
	abort.

replace_defs_form(F,Defs,Form) :-
	atom(F), !,
	(
	    member( F = Def_F , Defs) ->
	            subtract( Defs, [ F = Def_F], Defs1),  %% Defs1 = Defs \ { F = Def_F }
		    replace_defs_form( Def_F , Defs1, Form)
	      ;
	    Form = F
	).
replace_defs_form( #f , _Defs, #f) :- !.


replace_defs_form( ~F , Defs, ~Form) :- !,
	replace_defs_form( F, Defs, Form).

replace_defs_form( F1 & F2 , Defs, Form1 & Form2 ) :- !,
	replace_defs_form( F1, Defs, Form1),
	replace_defs_form( F2, Defs, Form2).


replace_defs_form( F1 v F2 , Defs, Form1 v Form2 ) :- !,
	replace_defs_form( F1, Defs, Form1),
	replace_defs_form( F2, Defs, Form2).

replace_defs_form( F1 --> F2 , Defs, Form1 --> Form2 ) :- !,
	replace_defs_form( F1, Defs, Form1),
	replace_defs_form( F2, Defs, Form2).


%%% ERROR

replace_defs_form(F,_Defs,_Form) :-
	print_error('Not a valid formula (check the syntax):\n'),
	print_error(' ~w\n',[F]),
	abort.























