%################# PRINT  UTILITIES  ######################%




:- module(printUtil, [
	         proof_to_latex/5,  % trees_to_latex(+FormLabel,+Form,+Proof,+Defs,+LatexFilePath)
		 generate_pdf/2,    % generate_pdf(+LatexFilePath,-PdfFilePath)
	         print_error/1,
	         print_error/2
              ]
           ).

:- use_module(basic, [
		  tr_formula/2,
		  contains_rule/2
	      ]
	     ).

:- use_module(syntax).

:- use_module(translationPhi, [phi/2] ).
:- use_module(translationTheta, [theta/2] ).
:- use_module(substElim, [
		  subst_elim/2
	      ] ).

preamble_latex_file('preamble.tex').  %%% File containing the preamble for the Latex file


%%%%%%%%     LATEX   RENDERING      %%%%%%%%%%%%%%%%%%%%%%%
%
%
%

newpage :-	nl,writeln('\\newpage'),nl.
section(Title) :-
	write('\\section*{'), write(Title),writeln('}').
medskip :- nl,writeln('\\medskip'),nl.

% proof_to_latex(+FormLabel,+Form,+MJCR_Proof,+Defs, +LatexFilePath)


proof_to_latex(FormLabel,Form,MJCR_Proof,Defs,LatexFilePath) :-
	tell(LatexFilePath), % open LatexFilePath
	tr_defs(Defs,DefsNF), %  DefsNF is Defs in normal form (no defined operators)
	write_preamble,
	write_test_label(FormLabel,Form),
	%%% Write MJCR derivation
	section('Derivation in $\\MJCR$'), medskip,
	write_definitions(DefsNF),
	write_proof(MJCR_Proof,DefsNF),
	%%% Write the MJC translation (with subst rule)
	newpage, section('Translation in $\\MJC$'), medskip,
	write_definitions(DefsNF),
	phi(MJCR_Proof,MJC_Proof), %% MJC translation
	write_proof(MJC_Proof,DefsNF),
	(
	    contains_rule(rOrp,MJC_Proof) ->
	     write_msg_rOrp
	    ;
	    true
	),
	(
	    %%% Eliminate subst rules from MJC_Proof, if any
	    contains_rule(subst,MJC_Proof) ->
	      newpage, section('Elimination of substitutions'), medskip,
	      write_definitions(DefsNF),
	      subst_elim(MJC_Proof,MJC_Proof_substFree),
	      write_proof(MJC_Proof_substFree,DefsNF),
	     (
		  contains_rule(rOrp,MJC_Proof) ->
		   write_msg_rOrp
		  ;
		  true
	      )
	    ;
	    %  MJC_Proof does not contain subst applications
	     MJC_Proof_substFree = MJC_Proof
	),
	%%% Write the NC translation
	newpage, section('Translation in $\\NC$'), medskip,
	write_definitions(DefsNF),
	theta(MJC_Proof_substFree,NC_Proof), %% NC translation
	write_proof(NC_Proof,DefsNF),
	(
	    contains_rule(orpI,NC_Proof) ->
	    write_msg_orpI
	    ;
	    true
	),
	nl,writeln('\\end{document}'),
	told. % close open files





proof_to_latex(_,_,_,_,_) :-  % ERRORS
	told, % close open files
	writeln('***  proof_to_latex errors   ***'),
	abort.


write_test_label(FormLabel,Form) :-
	writeln('\\noindent'),
	write('{\\bf Formula label:} \\verb|'),
	write(FormLabel),
	writeln('|'),
	writeln('\\['),
	write_formula(Form),
	writeln('\\]').


write_msg_rOrp :-
	writeln('\\medskip'),
	writeln('The rule \\['),
	writeln('\\prftree[r,doubleline]{$\\scriptstyle\\ruleRorp$}'),
	writeln('{\\prfassumption{\\seqmjc{\\fbot{(A\\lor B)},\\fbot{B},\\Gamma}{A}}}   '),
        writeln(' {\\seqmjc{\\Gamma}{A\\lor B}}  '),
	writeln('\\]'),
	writeln('can be derived in $\\MJC$ in few steps.').

write_msg_orpI :-
	writeln('\\medskip'),
	writeln('The rule \\['),
	writeln('\\prftree[r,doubleline]{$\\scriptstyle\\ruleOrpI$}'),
	writeln('{\\prfassumption{\\seqncup{\\fbot{(A\\lor B)},\\fbot{B},\\Gamma}{A}}}   '),
        writeln(' {\\seqncup{\\Gamma}{A\\lor B}}  '),
	writeln('\\]'),
	writeln('can be derived in $\\NC$ in few steps.').


%  generate_pdf(+LatexFilePath)

generate_pdf(LatexFilePath,PdfFilePath) :-
	file_name_extension(BaseName, tex, LatexFilePath),
	file_directory_name(LatexFilePath,Dir),
	% build command line to execute
	atomic_list_concat(['pdflatex -interaction=batchmode  -output-directory ', Dir,
			    ' ', LatexFilePath, '  > /dev/null'],
			   Pdflatex_cmd),
	shell(Pdflatex_cmd,Status), % exec Pdflatex_cmd, Status is the outcome

	(
	Status == 0 ->
	 %% pdflatex OK
	 file_name_extension(BaseName, pdf, PdfFilePath),
	  % clean
	   remove_files_suffList(Dir,[aux,log])
	;
	 % pdflatex errors
	  told, % close open files
	  print_error('The command line\n'),
	  print_error('  ~w\n',[Pdflatex_cmd] ),
	  print_error('failed. Check that the file prftree.sty is in the right directory\n'),
	  abort
	).

% remove_files_suff(+Dir,+Suff)
% remove all the files in Dir with suffix Suff

remove_files_suff(Dir,Suff) :-
	file_name_extension('*', Suff, StarSuff),
	directory_file_path(Dir,StarSuff, Dir_starSuff),
	expand_file_name(Dir_starSuff, Aux_list),
	forall( member(F,Aux_list), delete_file(F)).

%remove_files_suffList(+Dir,+SuffList)
% remove all the files in Dir with suffix in SuffList

remove_files_suffList(Dir,SuffList) :-
	forall(member(Suff,SuffList), remove_files_suff(Dir,Suff)).



%========================================================

%%     RULE NAMES

% NC RULES
%

write_rule(id) :-  !, write('\\ruleID').
write_rule(coerc) :-  !, write('\\coerc').
write_rule(boteCl) :-  !, write('\\botecl').
write_rule(markE) :-  !, write('\\markel').
write_rule(andI) :-  !, write('\\land I').
write_rule(and0E) :-  !, write('\\land E_0').
write_rule(and1E) :-  !, write('\\land E_1').
write_rule(or0I) :-  !, write('\\lor I_0').
write_rule(or1I) :-  !, write('\\lor I_1').
write_rule(orE) :-  !, write('\\lor E').
write_rule(implI) :-  !, write('\\imp I').
write_rule(implE) :-  !, write('\\imp E').

% derivable rules

write_rule(orpI) :-  !, write('\\ruleOrpI').

%   MJC RULES

write_rule(ax) :-  !, write('\\ruleAX').  % (*)
write_rule(lBot) :-  !, write('L\\bot').  % (*)
write_rule(lc0) :-  !, write('\\ruleLCZ').
write_rule(lc1) :-  !, write('\\ruleLCO').
write_rule(rc) :-  !, write('\\ruleRC').
write_rule(lAnd0) :-  !, write('L\\land_0').  % (*)
write_rule(lAnd1) :-  !, write('L\\land_1').  % (*)
write_rule(rAnd) :-  !, write('R\\land').   % (*)
write_rule(lOr) :-  !, write('L\\lor').  % (*)
write_rule(rOr0) :-  !, write('R\\lor_0').
write_rule(rOr1) :-  !, write('R\\lor_1').
write_rule(lImpl) :-  !, write('L\\imp').  % (*)
write_rule(rImpl) :-  !, write('R\\imp').  % (*)

% derivable rules

write_rule(lNeg) :-  !, write('\\ruleLNEG').
write_rule(rOrp) :-  !, write('\\ruleRorp').

% MJCR RULES
% all the rules of MJC marked by (*) and

write_rule(rest) :-  !, write('\\ruleRest').
write_rule(lf0) :-  !, write('\\ruleLFz').
write_rule(lf1) :-  !, write('\\ruleLFo').
write_rule(rlf) :-  !, write('\\ruleRLF').
write_rule(rOr) :-  !, write('R\\lor').



% ERROR: not defined Rule
write_rule(R) :- write('\\mbox{ !!!' ), write(R), write('}').


%##########   WRITE TREES ############



write_preamble :-
	preamble_latex_file(Preamble),  %% file containing the preamble
	(
	exists_file(Preamble) ->
	    open(Preamble,  read, Stream),
	    read_stream_to_codes(Stream,CodeList),
	    forall( member(Code, CodeList), (char_code(Char,Code), write(Char))),
	   close(Stream)
	;  %%% ERROR
	  told,
          print_error('File not found:\n'),
	  print_error('  ~w\n',[Preamble]),
	  print_error('This file should be in the mjc directory\n'),
	  abort
      ).




%%% Write formulas
% formulas not to be embraced by ()

is_atomic_or_negated_formula(A):- atom(A),!.
%is_atomic_or_negated_formula(A^):- atom(A),!.
is_atomic_or_negated_formula(#f ):- !.
is_atomic_or_negated_formula(_A --> #f) .
is_atomic_or_negated_formula(~ _A ) .

%%%  tr_defs(+Defs, -Defs1)
%
% Defs is a list of definitions of the kind
%     A =  D
%
%  Defs1 is obtained from D by replacing in D the defined operators
% (e.g., ~, <-->).
%
%  We say that Defs1 is in *normal form*

tr_defs([],[]).
tr_defs([ A = D | Defs], [A =  D1 | Defs1]) :-
	tr_formula(D,D1),
	tr_defs(Defs,Defs1).


% pretty_print_defFormula(+F,-F1)
% F is a  formula used as a definition, F1 its 'pretty print' version

pretty_print_defFormula(F,F1) :-
	upcase_atom(F,F1).




%%%  replace_defs(+F, +Defs, -F1)
%
% F1 is obtained by replacing every subformula in F with its definition
%   Defs is a list of definitions in normal form of the kind
%     A = D
% For instance, if
%
%  F     =  (a --> b v c) & (b v c --> d)
%  Defs  =  [ 'A' =  a --> b v c,  'B' = b v c]
%
%  we get F1 = 'A' & ('B' --> d)
%

replace_defs(F, Defs, A1) :- member( A = F,Defs), !,
	pretty_print_defFormula(A,A1).

replace_defs(F,Defs,F1) :-
	  F =.. [Op | Args ],
	  replace_defs_list(Args,Defs,Args1),
	  F1 =.. [Op| Args1 ].

%%%  replace_defs(+L, +Defs, -L1)
%  Apply replace_defs to all the formulas in L

replace_defs_list([],_Defs,[]).
replace_defs_list([F | Fs], Defs, [F1 |Fs1]):-
	replace_defs(F,Defs,F1),
	replace_defs_list(Fs,Defs,Fs1).


% write_formula(+F:formula,+Defs)
% Write the formula F using the definitions in Defs
% Defs is a list of definitions in normal form of the kind
%   A = D

write_formula(F,Defs) :-
	replace_defs(F,Defs,F1),
	write_formula(F1).

write_formula(#f) :- !,
        write('\\bot').

write_formula(A^) :- !,
	(
	 is_atomic_or_negated_formula(A) ->
            write('\\fbot{'),
             write_formula(A),
             write('}')
	;
	   write('\\fbot{('),
             write_formula(A),
             write(')}')
       ).

write_formula(F):-
        atom(F), !,
        write(F).

write_formula( A --> #f) :- !,
        (
         is_atomic_or_negated_formula(A) ->
           write('\\neg '),  write_formula(A)
        ;
          write('\\neg '), write('('),   write_formula(A), write(')')
        ).


write_formula( (A --> B) & (B-->A)) :- !,
         ( is_atomic_or_negated_formula(A) ->
            write_formula(A)
         ;
           write('('), write_formula(A), write(')')
        ),
         write('\\leftrightarrow '),
         ( is_atomic_or_negated_formula(B) ->
            write_formula(B)
         ;
           write('('), write_formula(B), write(')')
        ).

write_formula(F):-  % F is a unary formula
        F =.. [Op, A ], !,
        write_op(Op),
       ( is_atomic_or_negated_formula(A) ->
            write_formula(A)
         ;
           write('('), write_formula(A), write(')')
        ).

write_formula(F):-  % F is a binary formula
        F =.. [Op, A , B], !,
        ( is_atomic_or_negated_formula(A) ->
            write_formula(A)
         ;
           write('('), write_formula(A), write(')')
        ),
        write_op(Op),
        ( is_atomic_or_negated_formula(B) ->
            write_formula(B)
         ;
           write('('), write_formula(B), write(')')
        ).

%%%	ERROR
write_formula(F) :-
	told, % close open files
	writeln(F),
	writeln('***** not a formula ****'), abort.

write_op('~') :-  !, write(' \\neg ').
write_op('&') :- !, write(' \\land ').
write_op('v') :- !, write(' \\lor ').
write_op('-->') :-!, write(' \\imp ').
%%%
%%%	ERROR
write_op(Op) :-
	told, % close open files
	writeln(Op),
	writeln('***** not a logical connective ****'), abort.



%%% write_formula_list(+L,+Defs)
% Write the list of formulas L using definitions in Defs
% Defs: as above
%
write_formula_list([],_Defs):- !.
write_formula_list([A],Defs) :- write_formula(A,Defs), !.
write_formula_list([A | L],Defs) :-
	write_formula(A,Defs), write(',\\,'),
	write_formula_list(L,Defs).

%%% write_formula_list(+L, +Defs)
% Write the list of formulasL in lexicographic order
% Defs: as above

write_formula_list_ord(Formulas,Defs):-
	sort(Formulas,Formulas_ord),
	write_formula_list(Formulas_ord,Defs).


%%% write_definitions(+Defs)
%  Defs: list of definitions of the form
%   A = formula

write_definitions([]):- !.

write_definitions(Defs):-
	writeln('\\['),
	writeln('\\begin{array}{lclcl}'),
	write_defs_list(Defs,Defs),
	writeln('\\end{array}'),
	writeln('\\]').


write_def(A = F , Defs) :-
	pretty_print_defFormula(A,A1),
	write(A1),
	subtract(Defs, [A = F], Defs1), % Defs1 = Defs \{ A= F \}
	write(' & \\;=\\; & '),
	replace_defs(F,Defs1,F1),
	write_formula(F1,Defs1),
	(  F == F1 ->  % F does not contain definitions
	    true
	;   % write F without definitions
	write(' & \\;=\\; & '),
	write_formula(F,[])
	).

write_defs_list([D1],Ds) :-
	write_def(D1,Ds),
	nl.

write_defs_list([D1| Ds], Defs ) :-
	write_def(D1,Defs),
	nl,writeln('\\\\'),
	write_defs_list(Ds,Defs).


% big_union( +L , -Gamma)
%
%  L: list of lists of formula
%  Gamma =  union of the lists in L

%big_union([],[]).
%big_union([ I] , I).
%big_union([ I | L ], U) :-
%	big_union(L, V),
%	union(I,V,U).


% write_sequent(+Sequent, +Defs)
%%%   NC
%


write_sequent(Gamma '|--' up(A), Defs) :- !,
	write('\\seqncup{'),
	write_formula_list_ord(Gamma,Defs),
	write('}{'),
	write_formula(A,Defs),
	write('}').


write_sequent(Gamma '|--' down(A), Defs) :- !,
	write('\\seqncdown{'),
	write_formula_list_ord(Gamma,Defs),
	write('}{'),
	write_formula(A,Defs),
	write('}').




%%% MJCR


write_sequent(Gamma  ==> A @ Delta, Defs) :- !,
	write('\\seqmjcr{'),
	write_formula_list_ord(Gamma,Defs),
	write('}{'),
	write_formula(A,Defs),
	write('}{'),
	write_formula_list_ord(Delta,Defs),
	write('}').


write_sequent( Gamma @ Theta @ foc(H) =*=> lab(Label) @ F @ Delta , Defs  ) :- !,
	write('\\seqmjcrf{'),
	write(Label),
	write('}{'),
	write_formula_list_ord(Gamma,Defs),
	write('}{'),
	write_formula_list_ord(Theta,Defs),
	write('}{'),
	write_formula(H,Defs),
	write('}{'),
	write_formula(F,Defs),
	write('}{'),
	write_formula_list_ord(Delta,Defs),
	write('}').

%%% MJC
%%% ***  DO NOT MOVE!!! (problem with unification)

write_sequent(Gamma  ==> A,Defs ) :- !,
	write('\\seqmjc{'),
	%sort(Gamma,GammaOrd),
	write_formula_list_ord(Gamma,Defs),
	write('}{'),
	write_formula(A,Defs),
	write('}').


write_sequent( Gamma  @ foc(H) =*=> lab(Label) @ F , Defs ) :- !,
	write('\\seqmjcf{'),
	write(Label),
	write('}{'),
	write_formula_list_ord(Gamma,Defs),
	write('}{'),
	write_formula(H,Defs),
	write('}{'),
	write_formula(F,Defs),
	write('}').

% patterns


write_sequent( Gamma  @ foc(H) '|>' , Defs  ) :- !,
	write('\\seqmjcpat{'),
	write_formula_list_ord(Gamma, Defs),
	write('}{'),
	write_formula(H, Defs),
	write('}').


%%%
%%%	ERROR
write_sequent(E, _Defs) :-
	told, % close open files
	writeln(E),
	writeln('***** not a sequent ****'), abort.


% write_proof(+Tree, +Defs)
% write a proof
%
%  Use
%
%  flag(+Key, -Old, +New)
%
% to set a counter

write_proof(P,Defs) :- !,
	flag(count,_,0),
	writeln('\\['),
	write_proof_prf(P,Defs),
	writeln('\\]').


%% write_proof_prf(+Tree, +Defs)
% write a proof using prftree

write_proof_prf(proof(leaf,  Sequent,[]), Defs) :- !,
	write('\\prfassumption{ '),
	write_sequent(Sequent, Defs),
	writeln(' }').

write_proof_prf(proof(subst,  Sequent, Premise_list), Defs ) :- !,
	write('\\prftree[dottedline]'),
	write_proof_prf_premise_list(Premise_list, Defs),
	write('{ '),
	write_sequent(Sequent, Defs ),
	writeln(' }').


write_proof_prf(proof(Rule,  Sequent, Premise_list), Defs ) :- !,
        write('\\prftree[r]'),
	( member(Rule,[lNeg,negE,rOrp,orpI]) ->
		 write('[doubleline]')
		       ;
		 true
	),
	write('{$\\scriptstyle '),
	%Write('{$'),
	write_rule(Rule),
	writeln('$}'),
	write_proof_prf_premise_list(Premise_list, Defs),
	write('{ '),
	write_sequent(Sequent, Defs ),
	writeln(' }').

%%%	 ERROR
write_proof_prf(E,_Defs) :-
	told, % close open files
	writeln(E),
	writeln('***** not a proof ****'), abort.



write_proof_prf_premise_list([], _Defs):- !,
	writeln('{}').

write_proof_prf_premise_list([Tree], Defs ) :- !,
	flag(count,N, N+1),
	write('{%'),
	writeln(N),
	write_proof_prf(Tree,Defs),
	write('}%'),
	writeln(N).

write_proof_prf_premise_list([Tree | Tree_list], Defs) :-
	flag(count,N, N+1),
	write('{%'),
	writeln(N),
	write_proof_prf(Tree,Defs),
	write('}%'),
	writeln(N),
	write_proof_prf_premise_list(Tree_list,Defs).


%  print_error(+Format)
%  print_error(+Format,+Arguments)
%  Format: a Prolog atom
%  Arguments: a list


print_error(Format,Arguments) :-
	atomic_list_concat(['ERROR: ', Format], Error_Format),
	ansi_format([bold,fg(red)], Error_Format, Arguments).

print_error(Format) :-
     print_error(Format, []).





























